import numpy as np
import pandas as pd
import scipy.signal
from scipy import signal
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit, leastsq

from solarspatialtools import spatial


def correlation(baseline, estimation, scaling='coeff'):
    """
    Compute the cross correlation between two signals, including the full range
    of possible time lags.

    The value of correlation at lags==0 is the traditional cross-correlation
    without allowing for time shifting.

    This function is essentially a wrapper for scipy.stats.correlate, providing
    pre-baked scaling options, and always returning the lags.

    Parameters
    ----------
    baseline : numeric or pandas.Series
        The baseline signal

    estimation : numeric or pandas.Series
        The predicted signal

    scaling : str
        a string for the type scaling to use.

        Options are:
            `energy`:
                scales by the energy of the autocorrelation of the input
            `coeff`:
                scales the output to the correlation coefficient (always
                removes the mean from both signals)
            `unbiased_energy`:
                similar to energy, but normalizes based on lag to account for
                the fewer points used in the convolution. Removes bias towards
                small lags.
            `unbiased_coeff`:
                similar to coeff, but normalizes based on lag to
                account for the fewer points used in the
                convolution. Removes bias towards small lags.

    Returns
    -------
    corr : numeric
        A vector of the cross correlations

    lag : numeric
        A vector of the lag for each cross correlation
    """

    # The time lags
    lags = scipy.signal.correlation_lags(len(baseline), len(estimation))

    if scaling.lower() == 'energy':
        autocorr = scipy.signal.correlate(baseline, baseline)
        energy = np.max(autocorr)
        corr = scipy.signal.correlate(baseline, estimation) / energy
    elif scaling.lower() == 'coeff':
        num = scipy.signal.correlate(baseline - np.mean(baseline),
                                     estimation - np.mean(estimation))
        den = len(baseline) * (np.std(baseline) * np.std(estimation))
        corr = num / den
    elif scaling.lower() == 'unbiased_energy':
        autocorr = scipy.signal.correlate(baseline, baseline)
        energy = np.max(autocorr)
        corr = scipy.signal.correlate(baseline, estimation) / energy
        scale = (len(baseline) - np.abs(lags)) / len(baseline)
        corr /= scale
    elif scaling.lower() == 'unbiased_coeff':
        num = scipy.signal.correlate(baseline - np.mean(baseline),
                                     estimation - np.mean(estimation))
        den = len(baseline) * (np.std(baseline) * np.std(estimation))
        corr = num / den
        scale = (len(baseline) - np.abs(lags)) / len(baseline)
        corr /= scale
    elif scaling.lower() == 'none':
        corr = scipy.signal.correlate(baseline, estimation)
    else:
        raise ValueError("Illegal scaling specified.")

    return corr, lags


def _fftcorrelate(baseline, estimation, scaling='coeff'):
    """
    Compute the cross correlation between two signals, including the full range
    of possible time lags, using cross correlation. Can handle multiple time
    series.

    Parameters
    ----------
    baseline: pd.DataFrame
    estimation: pd.DataFrame
    scaling: str (default 'coeff')
        Type of scaling to use for cross correlation. Note that debiased
        scaling options are not available here. Valid options are:
        'coeff' - computes the correlation coefficient
        'none' - no scaling is applied

    Returns
    -------
    corr : numeric
        A vector of the cross correlations for each pairing in the inputs

    lag : numeric
        A vector of the lag for each cross correlation
    """

    # Compute the lags
    lags = signal.correlation_lags(len(baseline), len(estimation))

    # Rename for brevity
    ts_inm = baseline
    ts_outm = estimation

    # Condition pandas input types
    if isinstance(ts_inm, pd.DataFrame):
        ts_inm = np.array(ts_inm).T
    if isinstance(ts_outm, pd.DataFrame):
        ts_outm = np.array(ts_outm).T

    # Condition vector inputs
    if ts_inm.ndim == 1:
        ts_inm = np.expand_dims(ts_inm, axis=0)
    if ts_outm.ndim == 1:
        ts_outm = np.expand_dims(ts_outm, axis=0)

    # Perform the scaling calculations
    if scaling == 'coeff':
        # Subtract means
        ts_inm = ts_inm - np.expand_dims(np.mean(ts_inm, axis=1), axis=1)
        ts_outm = ts_outm - np.expand_dims(np.mean(ts_outm, axis=1), axis=1)
        # Compute the scaling factor (normalizing by stdev)
        corr_scale = (np.size(ts_inm, axis=1)
                      * np.std(ts_inm, axis=1)
                      * np.std(ts_outm, axis=1)) ** -1
    elif scaling == 'none':
        corr_scale = 1
    else:
        raise ValueError(f"Illegal scaling specified: {scaling}.")

    na = np.size(ts_inm, axis=1)
    nl = 2 * na - 1
    addon = np.zeros((ts_inm.shape[0], nl - na))
    ts_inm = np.concatenate([ts_inm, addon], axis=1)
    ts_outm = np.concatenate([ts_outm, addon], axis=1)

    # Compute correlation via fft and re-center to match lags
    ffta = scipy.fft.fft(ts_inm, axis=1)
    fftb = scipy.fft.fft(ts_outm, axis=1)
    corrxy = np.real(np.fft.ifft(ffta * np.conj(fftb), axis=1)) * corr_scale

    corrxy = np.roll(corrxy, nl // 2, axis=1)

    return corrxy, lags


def averaged_psd(input_tsig, navgs, overlap=0.5,
                 window='hamming', detrend='linear', scaling='density'):
    """
    Calculate an averaged power spectral density for a signal.

    Parameters
    ----------
    input_tsig : numeric
        Pandas type with the TF input time signal. Index must be time.

    navgs : int
        The number of averages to use based on zero overlap. Overlap will
        result in more averages.

    overlap : float (default 0.5)
        Percentage overlap between the averages

    window : string (default 'hamming')
        The window type to use.

    detrend : string
        Detrend type ('linear' or 'constant'). See scipy.signal.welch for more
        information.

    scaling : string (default 'density')
        The type of scaling to request from scipy. See scipy.signal.welch for
        more info

    Returns
    -------
    output : Series
        Pandas Series containing the power spectral density with an index of
        the frequency.
    """
    dt = (input_tsig.index[1] - input_tsig.index[0]).total_seconds()
    fs = 1/dt

    if isinstance(input_tsig, pd.Series):
        return averaged_psd(pd.DataFrame(input_tsig), navgs, overlap, window,
                            detrend)

    cols = input_tsig.columns
    input_tsig = input_tsig.values.T

    nperseg = int(input_tsig.shape[1] // navgs)
    noverlap = int(nperseg * overlap)

    f, psdxx = signal.welch(input_tsig, fs=fs, window=window,
                            nperseg=nperseg, detrend=detrend,
                            noverlap=noverlap, scaling=scaling)
    # Reported units from scipy are V**2/Hz
    return pd.DataFrame(psdxx.T, index=f, columns=cols)


def averaged_tf(input_tsig, output_tsig,
                navgs, overlap=0.5, window='hamming', detrend='linear'):
    """
    Calculate a transfer function between two signals, along with their
    coherence.

    Parameters
    ----------
    input_tsig : numeric
        Pandas type with the TF input time signal. Index must be time.

    output_tsig : numeric
        Pandas type with the TF output time signal. Index must be time.

    navgs : int
        The number of averages to use based on zero overlap. Overlap will
        result in more averages.

    overlap : float (default 0.5)
        Percentage overlap between the averages

    window : string (default 'hamming')
        The window type to use.

    detrend : string or None
        Detrend type ('linear' or 'constant'). See scipy.signal.psdxx for more
        information.

    Returns
    -------
    tf : pd.DataFrame
        Complex valued transfer function indexed by the frequency and
        containing the same columns as the input
    coh : pd.DataFrame
        the coherence indexed by the frequency and containing the same
        columns as the input
    """
    dt = (input_tsig.index[1] - input_tsig.index[0]).total_seconds()
    fs = 1/dt

    if isinstance(output_tsig, pd.Series):
        return averaged_tf(input_tsig, pd.DataFrame(output_tsig),
                           navgs, overlap, window, detrend)
    if isinstance(input_tsig, pd.Series):
        return averaged_tf(pd.DataFrame(input_tsig), output_tsig, navgs,
                           overlap, window, detrend)

    if not (isinstance(output_tsig, pd.DataFrame)
            and isinstance(input_tsig, pd.DataFrame)):
        raise ValueError('Input and output timesignals must be pandas types')

    out_cols = output_tsig.columns
    in_cols = input_tsig.columns
    output_tsig = output_tsig.values.T
    input_tsig = input_tsig.values.T

    if len(out_cols) == len(in_cols):
        cols = [f"{in_cols[i]}_{out_cols[i]}" for i in range(len(out_cols))]
    elif len(out_cols) == 1:
        cols = [f"{in_cols[i]}_{out_cols[0]}" for i in range(len(in_cols))]
    elif len(in_cols) == 1:
        cols = [f"{in_cols[0]}_{out_cols[i]}" for i in range(len(out_cols))]
    else:
        raise RuntimeError("Shouldn't Be Possible")

    nperseg = int(input_tsig.shape[1] // navgs)
    noverlap = int(nperseg * overlap)

    # Calculate the transfer function
    _, psdxx = signal.welch(input_tsig, fs=fs, window=window,
                            nperseg=nperseg, detrend=detrend,
                            noverlap=noverlap, scaling='density')
    _, psdyy = signal.welch(output_tsig, fs=fs, window=window,
                            nperseg=nperseg, detrend=detrend,
                            noverlap=noverlap, scaling='density')
    freqs, csdxy = signal.csd(input_tsig, output_tsig, fs=fs, window=window,
                              nperseg=nperseg, detrend=detrend,
                              noverlap=noverlap)

    # Calculating manually allows for vectorization and is faster
    with np.errstate(divide='ignore', invalid='ignore'):
        tf = csdxy / psdxx
        coh = np.abs(csdxy) ** 2 / (psdxx * psdyy)

    # Calculate the coherence
    # _, coh = signal.coherence(input_tsig, output_tsig, fs=fs, window=window,
    #                           nperseg=nperseg, noverlap=noverlap,
    #                           detrend=detrend)

    tf = pd.DataFrame(tf.T, index=freqs, columns=cols)
    coh = pd.DataFrame(coh.T, index=freqs, columns=cols)

    return tf, coh


def interp_tf(new_freq, input_tf):
    """
    Interpolate a transfer function in the frequency domain by magnitude and
    phase independently. This is necessary because the complex interpolation
    doesn't really do the job on its own.

    Parameters
    ----------
    new_freq : np.array or pd.Index
        The new frequency index to interpolate onto.

    input_tf : pd.Series or pd.DataFrame
        The transfer function to be interpolated

    Returns
    -------
    interp_tf : pd.Series or pd.DataFrame
        The transfer function interpolated to the new frequency axis. Type will
        match the type of the input.
    """
    sortinds = input_tf.index.argsort()
    if type(input_tf) is type(pd.Series()):
        use_tf = pd.DataFrame(input_tf)
    else:
        use_tf = input_tf

    # Generate a function handle and interpolate the magnitude
    interp_mag_func = interp1d(use_tf.index[sortinds],
                               np.abs(use_tf.iloc[sortinds, :]),
                               axis=0)
    interp_mag = interp_mag_func(new_freq)

    # Generate a function handle and interpolate the phase
    # Work on the unwrapped angle to make sure that we don't have weird
    # results in the middle of wraps.
    interp_phase_func = interp1d(use_tf.index[sortinds],
                                 np.unwrap(np.angle(use_tf.iloc[sortinds, :]),
                                           axis=0),
                                 axis=0)
    interp_phase = interp_phase_func(new_freq)

    # Recreate the complex TF
    interp_filt = interp_mag * np.exp(1j * interp_phase)

    # Appropriately recast the type
    if type(input_tf) is type(pd.Series()):
        interp_filt = pd.Series(interp_filt[:, 0], index=new_freq)
    else:
        interp_filt = pd.DataFrame(interp_filt, columns=input_tf.columns,
                                   index=new_freq)
    return interp_filt


def tf_delay(tf, coh, coh_limit=0.6, freq_limit=0.02, method='fit'):
    """
    Compute the delay based on the phase of a transfer function

    Parameters
    ----------
    tf : pd.DataFrame
        The transfer functions, produced by signalproc.averaged_tf.
    coh : pd.DataFrame
        The coherence, produced by signalproc.averaged_tf. Must match shape
        of tf.
    coh_limit : float
        The coherence limit to use for filtering the phase. Set to None to
        include all points.
    freq_limit : float
        The frequency limit to use for filtering the phase. Set to None to
        include all points.
    method : str
        The method to use for computing the delay.

        Options are:
            'diff' - unwrap phase and take derivative (not well validated)
            'fit' - fit a line to the phase
            'multi' - fit a line to the phase for multiple tfs at once

    Returns
    -------
    delay : float
        The delay in seconds
    ix : np.array(bool)
        The mask used for filtering the phase
    """

    def _delay_fitter(x, delval):
        """
        Helper function for use with curve_fit. Computes the modeled phase.
        Parameters
        ----------
        x :
            the transfer function frequency
        delval :
            The 'real' phase to fit the group delay to
        Returns
        -------
        model :
            the modeled phase
        """
        # model = np.unwrap(np.angle(np.ones_like(x) *
        #                            np.exp(2 * np.pi * 1j * x * -delval)))
        # Simplifies to a linear relationship
        model = 2 * np.pi * x * -delval
        return model

    def _residual(dels, freqs, phases, mask):
        """
        Helper function for use with leastsq. Computes the residual between
        the modeled phase and the measured phase.
        Parameters
        ----------
        dels : float
            The array of delay to use for the model
        freqs : float
            The frequency array (must be multidimensional)
        phases : float
            The measured phase array (must be multidimensional)
        mask : bool
            The mask to use for filtering the phase

        Returns
        -------
        residual : float
            The residual between the model and the measured phase
        """
        # model = np.unwrap(np.angle(np.ones_like(freqs) *
        #                            np.exp(2 * np.pi * 1j * freqs * -dels)),
        #                   axis=0)
        # Simplifies to a linear relationship
        model = 2*np.pi*freqs*-dels
        diff = (phases - model) * mask
        return np.nansum(diff, axis=0)

    if isinstance(tf, pd.Series):
        return tf_delay(pd.DataFrame(tf), coh, coh_limit, freq_limit, method)
    if isinstance(coh, pd.Series):
        return tf_delay(tf, pd.DataFrame(coh), coh_limit, freq_limit, method)

    if coh_limit is None:
        ix_coh = np.ones_like(coh, dtype=bool)
    else:
        ix_coh = (coh.values > coh_limit)
    if freq_limit is None:
        ix_freq = np.ones_like(tf, dtype=bool)
    else:
        ix_freq = np.zeros_like(tf, dtype=bool)
        ix_freq[tf.index < freq_limit, :] = True
    ix = np.bitwise_and(ix_coh, ix_freq)

    if method == 'diff':
        if not np.any(np.array(np.shape(tf)) == 1):
            raise ValueError('tf must be a 1D array for method: diff')

        # # Method 1: unwrap phase and take derivative
        tfv = tf.values.flatten()
        gd = -np.diff(np.unwrap(np.angle(tfv))) / np.diff(tf.index)
        gd = np.append(gd, gd[-1])
        gd = gd/(2*np.pi)
        avg_del = np.sum(gd * ix.flatten() / np.sum(ix))
        return avg_del, ix

    elif method == 'fit':
        # Method 2: curve fit the phase
        if not np.any(np.array(np.shape(tf)) == 1):
            raise ValueError('tf must be a 1D array for method: fit')
        try:
            return (curve_fit(_delay_fitter,
                              np.expand_dims(tf.index, axis=1)[ix],
                              np.unwrap(np.angle(tf.loc[ix]).flatten()))[0].item(),
                    ix)
        except ValueError:
            from warnings import warn
            if not ix.any():
                warn('Curve fit failed due to coherence limit. Returning NaN')
                return np.nan, ix

            else:
                warn('Curve fit failed for unknown reason. Returning NaN')
                return np.nan, ix
    elif method == "multi":
        # Method 3 : Operate on multiple tfs at a time using fit method

        # The guess for the delays
        guess = np.zeros((1, np.size(tf, axis=1)))

        # Get subsets on the frequency axis to simplify fitting
        ix_freq1 = ix_freq[:, 0]
        if isinstance(tf, (pd.DataFrame, pd.Series)):
            tf_subset = tf.values[ix_freq1, :]
        else:
            tf_subset = tf[ix_freq1, :]

        # The arguments for the residual function
        args = (np.expand_dims(tf.index[ix_freq1], axis=1),
                np.unwrap(np.angle(tf_subset), axis=0),
                ix_coh[ix_freq1, :])
        # Perform the simultaneous least squares fit
        best, cov = leastsq(_residual, guess, args=args)

        best[np.all(np.isnan(tf), axis=0)] = np.nan
        return best, ix
    else:
        raise ValueError(f'Invalid method: {method}')


def xcorr_delay(ts_in, ts_out, scaling='coeff'):
    """
    Compute the delay between two timeseries using cross correlation.

    Parameters
    ----------
    ts_in : pd.Series
        The input timeseries. Requires that the index operate as datetime.
    ts_out : pd.Series
        The output timeseries
    scaling : str
        Type of scaling to use for cross correlation.

        Options are:
            `energy`:
                scales by the energy of the autocorrelation of the input
            `coeff`:
                scales the output to the correlation coefficient (always
                removes the mean from both signals)
            `unbiased_energy`:
                similar to energy, but normalizes based on lag to account for
                the fewer points used in the convolution. Removes bias towards
                small lags.
            `unbiased_coeff`:
                similar to coeff, but normalizes based on lag to account for
                the fewer points used in the convolution. Removes bias towards
                small lags.

    Returns
    -------
    delay : float
        Time lag between the two timeseries at the maximum value of the cross
        correlation. Values are always integer multiples of the sampling period
        as the max correlation values are limited to the discrete time steps.
    corr : float
        The peak value of the cross correlation at the identified delay.
    """
    # Method 3: Cross Correlation
    xcorr_i, lags = correlation(ts_in, ts_out, scaling)
    dt = (ts_in.index[1] - ts_in.index[0]).total_seconds()
    lags = lags * dt
    peak_lag_index = xcorr_i.argmax()  # Index of peak correlation
    delay = -lags[peak_lag_index]
    return delay, xcorr_i[peak_lag_index]


def compute_delays(ts_in, ts_out, mode='loop', scaling='coeff'):
    """
    Compute the delay between two sets of timeseries using cross correlation.
    Uses similar methodologies to xcorr_delay, but allows for multiple input
    timeseries to be computed simultaneously.

    Parameters
    ----------
    ts_in: pd.DataFrame
        The time series representing the input signal
    ts_out: pd.DataFrame
        The time series representing the output signal
    mode: str (default 'loop')
        The method to use for computing the delay.

        Options are:
            `loop`:
                loop over each input timeseries pair and compute the xcorr
            `fft`:
                use the fft to compute the xcorr (seems to be slower)

    scaling: str (default 'coeff')
        Type of scaling to use for cross correlation. Note that debiased
        scaling options are not available here.

        Valid options are:
            `coeff`:
                computes the correlation coefficient
            `none`:
                no scaling is applied

    Returns
    -------
    delay : array(float)
        Time lag between the two timeseries at the maximum value of the cross
        correlation. Values are always integer multiples of the sampling period
        as the max correlation values are limited to the discrete time steps.
    extra_data : dict{}
        Additional data about the correlations. Values are:
        'peak_corr' - the peak value of the correlation for each input ts
        'mean_corr' - the mean correlation for each input timeseries
        'zero_corr' - the correlation at zero_lag for each input timeseries
    """
    lags = signal.correlation_lags(len(ts_in), len(ts_in))
    dt = (ts_in.index[1] - ts_in.index[0]).total_seconds()
    lags = lags * dt

    ts_inm = ts_in
    ts_outm = ts_out
    if isinstance(ts_in, pd.DataFrame):
        ts_inm = np.array(ts_inm).T
    if isinstance(ts_out, pd.DataFrame):
        ts_outm = np.array(ts_outm).T

    extras = {'peak_corr': np.zeros(ts_inm.shape[0]),
              'mean_corr': np.zeros(ts_inm.shape[0]),
              'zero_corr': np.zeros(ts_inm.shape[0])
              }
    zero_lag_ind = np.where(lags == 0)

    # Perform the scaling calculations
    if scaling == 'coeff':
        # Subtract means
        ts_inm = ts_inm - np.expand_dims(np.mean(ts_inm, axis=1), axis=1)
        ts_outm = ts_outm - np.expand_dims(np.mean(ts_outm, axis=1), axis=1)
        # Compute the scaling factor (normalizing by stdev)
        corr_scale = (np.size(ts_inm, axis=1)
                      * np.std(ts_inm, axis=1)
                      * np.std(ts_outm, axis=1)) ** -1
    elif scaling == 'none':
        corr_scale = 1
    else:
        raise ValueError(f"Illegal scaling specified: {scaling}.")

    if mode == 'loop':
        # signal.correlate can't handle 2D inputs without performing 2D conv.

        # Precomputing method saves time
        method = signal.choose_conv_method(ts_inm[0, :], ts_outm[0, :])

        # Preallocate and Compute for individual pairs
        delay = np.zeros(ts_inm.shape[0])
        for i in range(ts_outm.shape[0]):
            xcorr = signal.correlate(ts_inm[i, :], ts_outm[i, :],
                                     method=method)
            # xcorr, _ = correlation(ts_inm[i, :], ts_outm[i, :], 'none')
            peak_lag_index = xcorr.argmax()
            delay[i] = -lags[peak_lag_index]

            extras['peak_corr'][i] = xcorr[peak_lag_index]
            extras['mean_corr'][i] = np.mean(xcorr)
            extras['zero_corr'][i] = xcorr[zero_lag_ind].item()

        # Scale
        extras['peak_corr'] *= corr_scale
        extras['mean_corr'] *= corr_scale
        extras['zero_corr'] *= corr_scale

    elif mode == 'fft':
        # Have to pad with zeros to get equivalent vals to normal xcorr.
        na = np.size(ts_inm, axis=1)
        nl = 2*na - 1
        addon = np.zeros((ts_inm.shape[0], nl - na))
        ts_inm = np.concatenate([ts_inm, addon], axis=1)
        ts_outm = np.concatenate([ts_outm, addon], axis=1)

        # Need to recompute the lags to accommodate the shifted zero for fft
        lags = np.roll(lags, -nl // 2 + 1)
        zero_lag_ind = np.where(lags == 0)

        # Compute correlation via fft and re-center to match lags
        ffta = scipy.fft.fft(ts_inm, axis=1)
        fftb = scipy.fft.fft(ts_outm, axis=1)
        corrxy = np.real(np.fft.ifft(ffta * np.conj(fftb), axis=1))

        # Extract value and lag of peak correlation
        peak_lag_indices = corrxy.argmax(axis=1)
        delay = -lags[peak_lag_indices]

        extras['peak_corr'] = (corrxy[range(ts_inm.shape[0]), peak_lag_indices]
                               * corr_scale)
        extras['mean_corr'] = np.mean(corrxy, axis=1) * corr_scale
        extras['zero_corr'] = corrxy[:, zero_lag_ind] * corr_scale
    else:
        raise ValueError(f'Invalid mode: {mode}')

    # Sometimes produce erroneous delay for uncorrelated values
    delay[extras['peak_corr'] < 1e-10] = 0

    return delay, extras


def apply_delay(tf, delay):
    """
    Apply a time delay to a transfer function. This is equivalent to rotating
    the phase as a linear function of frequency.

    To compare with a unity transfer function, consider:
        apply_delay(tf * 0 + 1, delay)

    Parameters
    ----------
    tf : pd.DataFrame
        The transfer function, produced by signalproc.averaged_tf. The index
        must contain the frequency in Hz.
    delay : float
        The delay to apply to the transfer function. Units are in seconds.

    Returns
    -------
    tf : pd.DataFrame
        A copy of the transfer function with the delay applied in rotating
        the phase.
    """

    tfi = tf.copy().values.flatten()

    # Equation for rotating the phase by a delay
    tf = tfi * np.exp(2 * np.pi * 1j * tf.index * -delay)

    return tf


def get_1d_plant(centers, ref_center=0,
                 width=None, shape="square",
                 dx=1, xmax=500000):
    """
    Generate a one dimensional plant array based on a list of center positions.
    Plant is essentially a comb filter with a site of a given shape placed at
    each specified center position

    Parameters
    ----------
    centers : numeric
        List of centers of the individual measurement locations. Commonly the
        output of spatial.project_vectors().

    ref_center : numeric
        Position of the reference, will be used as the zero of the x coordinate

    width : numeric
        The size of each individual plant component. If None, is equivalent to
        the dx for the plant.

    shape : string
        The shape to use for each individual plant component. Choices are:
        'square', 'triangle', 'gaussian'

    dx : numeric
        The x axis spacing to use for the numerical plant layout.

    xmax : numeric
        The maximum x size to use for the plant domain

    Returns
    -------
    plant : numeric
        A vector representing the plant's density of generation along the x
        axis

    x_vec : numeric
        The position axis for the plant.
    """

    if width is None:
        w = dx
    else:
        w = width

    centers = np.array(centers).flatten()
    centers -= ref_center

    # Initialize the empty plant
    x_vec = np.arange(-xmax//2, xmax//2, dx)
    plant = np.zeros(x_vec.shape, dtype=float)

    # Creating the individual plant windows ###############

    if shape.lower() == "square":
        # Square individual plants
        # Lc = L total plant
        # north = # of sites
        # Lc/north = separation
        for center in centers:
            inds = np.bitwise_and(x_vec >= (center - w / 2),
                                  x_vec < (center + w / 2))
            plant[inds] = 1
    elif shape.lower() == "triangle":
        for center in centers:
            inds = np.bitwise_and(x_vec >= (center - w / 2),
                                  x_vec < (center + w / 2))
            plant[inds] = x_vec[inds] - center+w/2
    elif shape.lower() == "gaussian":
        # Gaussian Window
        for center in centers:
            plant += np.exp(-(x_vec-center)**2/(2*(w/2.355)**2))  # FWHM is STD
    else:
        raise ValueError("No info for plant shape: {}".format(shape))

    return plant, x_vec


def plant1d_to_camfilter(plant, x_plant, cloud_speed):
    """
    Take a 1D plant and compute the Cloud Advection Model representation

    Parameters
    ----------
    cloud_speed : numeric
        The cloud motion vector speed

    plant : np.array
        An array-based representation of the plant generation density. Will be
        normalized to produce a transfer function DC magnitude of 1. See
        get_1d_plant().

    x_plant : np.array
        The plant's x-coordinate. Should have a value of zero at the location
        of the reference point. See get_1d_plant().

    Returns
    -------
    filter : pd.Series
        A pandas series with the complex valued transfer function, indexed by
        the corresponding frequency.
    """
    # TODO needs to be validated

    dx = x_plant[1]-x_plant[0]

    plant = plant / np.sum(plant)  # normalize the plant
    camfilt = np.fft.fft(plant)  # What does it look like in f domain
    spatialdt = dx / np.abs(cloud_speed)  # Effective dt for cloud motion
    camfreq = np.fft.fftfreq(plant.shape[-1], spatialdt)

    # Shift the phase
    t_delay = np.min(x_plant) / cloud_speed
    if cloud_speed > 0:
        camfilt = camfilt * np.exp(
            1j * camfreq * (2 * np.pi) * t_delay)
    else:
        camfilt = np.conj(
            camfilt * np.exp(1j * camfreq * (2 * np.pi) * -t_delay))
    return pd.Series(camfilt, index=camfreq)


def apply_filter(input_tsig, comp_filt):
    """
    Apply a filter to a signal, and return the filtered signal. Works to align
    the frequency axis of the computed filter with the input signal.

    Parameters
    ----------
    input_tsig : pandas.Series or DataFrame
        Pandas type that contains the time signal

    comp_filt : Series, DataFrame
        Pandas type containing the complex valued filter to apply with its
        frequency in the index. See for example: get_camfilter

    Returns
    -------
    filtered_sig : numeric
        The filtered time series.
    """
    # Get the fft of the input signal, including its frequency axis
    dt = (input_tsig.index[1] - input_tsig.index[0]).total_seconds()
    input_fft = np.fft.fft(input_tsig) * 2 / len(input_tsig)
    f_vec = np.fft.fftfreq(input_tsig.shape[-1], dt)

    if np.max(f_vec) > np.max(comp_filt.index):
        raise ValueError('Error: the TF to apply does not cover the entire '
                         'frequency axis needed for the signal. Please '
                         'provide a TF with a higher maximum frequency.')

    # Interpolate the computational
    interp_filt = interp_tf(f_vec, comp_filt)

    # Apply the filter and invert.
    filtered_fft = input_fft * interp_filt
    filtered_sig = np.fft.ifft(filtered_fft * len(input_tsig) / 2)
    filtered_sig = np.real(filtered_sig)
    filtered_sig = pd.Series(filtered_sig, index=input_tsig.index)

    return filtered_sig


def get_camfilter(positions, cloud_speed, cloud_dir, ref_position, dx=1,
                  **kwargs):
    """
    Compute the filter for the CAM model

    Parameters
    ----------
    positions : pandas.DataFrame
        Pandas object containing locations of each reference site within the
        overall plant. Must be indexed by the site id. See data storage format.

        If positions contain 'lat' and 'lon' columns, they will be converted
        to UTM assuming latitude and longitude in degrees. Otherwise, it will
        be assumed that they are already in a UTM-like coordinate system.

    cloud_speed : numeric
        The cloud motion speed

    cloud_dir : tuple
        A tuple (dx,dy) representing the cloud motion direction. Will be
        converted to a unit vector, so length is not important.

    ref_position : pandas.DataFrame
        A subset of positions that represents the position of the reference.

    dx : numeric
        The spatial spacing that should be used in representing the plant.
        Affects the frequency band that can be represented.

    **kwargs : various
        Parameters that will be passed to get_1D_plant(). Include
            'width' - numeric width of each centered object
            'shape' - shape of each centered object (e.g. 'square')
            'xmax' - numeric maximum value in the spatial domain for the plant.
                     Affects the frequency resolution of the filter.
    Returns
    -------
    camfilter : Series
        A pandas Series containing the complex valued filter, along with its
        frequency vector along the index.
    """
    try:
        pos_utm = spatial.latlon2utm(positions['lat'], positions['lon'])
    except KeyError:
        pos_utm = positions
    try:
        ref_utm = spatial.latlon2utm(ref_position['lat'], ref_position['lon'])
        ref_utm = pd.Series(ref_utm, index=['E', 'N', 'zone'])
    except KeyError:
        ref_utm = ref_position

    pos_vecs = spatial.compute_vectors(pos_utm['E'], pos_utm['N'],
                                       ref_utm[['E', 'N']])
    pos_dists = spatial.project_vectors(pos_vecs, cloud_dir)

    plant, x_plant = get_1d_plant(pos_dists, dx=dx, **kwargs)
    camfilter = plant1d_to_camfilter(plant, x_plant, cloud_speed)
    return camfilter


def get_marcosfilter(s, freq=None):
    """
    Compute the filter for the Marcos model

    Parameters
    ----------
    s : numeric
        plant size in Hectares

    freq : numeric (default None)
        A vector of frequencies to include. A reference array will be computed
        if no frequency is provided.

    Returns
    -------
    output : Series
        A pandas Series with the complex valued filter. Index is frequency.

    """
    if freq is None:
        freq = np.linspace(0, 0.5, 100)
    k = 1
    fc = 0.02 / np.sqrt(s)
    filt = k / (1j * freq / fc + 1)
    return pd.Series(filt, index=freq, dtype=np.complex64)


def cleanfreq(sig):
    """
    Cleanup the bidirectional frequencies of a filter object for better
    visualization without lines wrapping across the zero.

    Parameters
    ----------
    sig : pandas.Series
        An object with an index of frequency that will be adjusted

    Returns
    -------
    The signal object with modified frequency
    """
    idxlist = sig.index.to_list()
    idxlist[len(sig.index) // 2] = None
    sig.index = idxlist



import warnings

import numpy as np
import pandas as pd

from solarspatialtools import spatial, cmv, signalproc
from scipy.optimize import linear_sum_assignment

warnings.filterwarnings("ignore",
                        message=".*Covariance of the parameters.*")


def compute_predicted_position(dfs, pos_utm, ref, cld_vecs=None,
                               mode='coherence', ndownsel=8,
                               navgs=5, coh_limit=0.6, freq_limit=0.02,
                               delay_method="multi"):
    """
    Compute the predicted position of a combiner based upon the cloud movement.
    Requires two separate inputs with different CMV directions. Inputs are
    provided as lists containing the data about the two different cases. For
    reference on the method, refer to [1]_.

    Parameters
    ----------
    dfs : list[pandas.DataFrame]
        Time series of the data. Columns are the points, rows are the time.
    pos_utm : pandas.DataFrame
        The known positions of all sensors. Must be in UTM (x,y) coordinates.
        Index should match the columns of df
    ref : str
        The name of the reference point. Must be a column in df and a row in
        pos_utm.
    cld_vecs : np.ndarray
        The cloud motion vectors, [[Vx1, Vy1], [Vx2, Vy2]]
    mode : str
        The method for downselecting the points to use for computing the
        position. Options are 'coherence', 'global_coherence', 'distance', and
        'all'. 'coherence' will choose the n best points for each CMV that
        experience the best coherence. 'global_coherence' computes an
        orthogonal sum of coherence for both CMV's and chooses the same best n
        points across both. Default is 'coherence'.
    ndownsel : int
        The number of points to downselect to. Default is 8.
    navgs : int
        The number of averages to use when computing the transfer function.
        Affects both the coherence and the frequency resolution. Default is 5.
        see solarspatialtools.signalproc.averaged_tf for more information.
    coh_limit : float
        The minimum coherence required for computing the delay. Default is 0.6.
        See solarspatialtools.signalproc.tf_delay for more information.
    freq_limit : float
        The maximum frequency that will be used when computing the delay.
        See solarspatialtools.signalproc.tf_delay for more info.
    delay_method : str
        The method for computing the delay. Options are 'fit' and 'multi'.

    Returns
    -------
    pos : np.ndarray
        The predicted position of the combiner (X,Y) in the coordinate system
        of pos_utm
    fulldata : list[pandas.DataFrame]
        The full data set for each of the two cases. Each element of the list
        is a pandas.DataFrame with columns 'cloud_dist', 'delay', and 'coh'
        and an index matching pos_utm. Cloud_dist is the distances implied by
        the cloud motion for that particular time along the cloud motion dir.
        Delay is the raw value of tf delay for each of the points relative to
        the reference. coh is the average coherence for that transfer function.

    References
    ----------
    .. [1] Ranalli, J., Hobbs, W., 2024. PV Plant Equipment Labels and
       Layouts Can Be Validated by Analyzing Cloud Motion in Existing Plant
       Measurements. IEEE Journal of Photovoltaics.
       DOI: `https://doi.org/10.1109/JPHOTOV.2024.3366666`
    """

    # Note: This approach computes results for a single reference. The sub-
    # function compute_cloud_dist needs the tf between every single location
    # and that reference point. Thus, when computing for a whole field, we
    # duplicate the work of computing the tf for every point pair. Efficiency
    # could be improved on plant-level calculations by pre-computing all delays
    # for every point pair.

    combined_data = []

    if cld_vecs is None:
        cld_vecs = []
        for df in dfs:
            cld_spd, cld_dir, dat = cmv.compute_cmv(df, pos_utm,
                                                    method='jamaly')
            cld_vecs.append(spatial.pol2rect(cld_spd, cld_dir))

    for df, cld_vec in zip(dfs, cld_vecs):

        # Get the pairwise delays
        delay, coh = compute_delays(df, ref, navgs=navgs,
                                    coh_limit=coh_limit,
                                    freq_limit=freq_limit,
                                    method=delay_method)

        # Convert them to distances along the cloud motion vector
        delay_dist = -delay * spatial.magnitude(cld_vec)

        # Compute the actual separation distance for comparison
        # Negative gives dist from all points to ref, rather than vice versa
        pos_vecs = -spatial.compute_vectors(pos_utm['E'], pos_utm['N'],
                                            pos_utm.loc[ref][['E', 'N']])
        # Project the vec from point to ref into the cloud vector dir
        plan_dist = spatial.project_vectors(pos_vecs, cld_vec)

        error_dist = delay_dist - plan_dist['dist']

        # Aggregate
        pos_data = pd.DataFrame({'plan_dist': plan_dist['dist'],
                                 'delay_dist': delay_dist,
                                 'error_dist': error_dist,
                                 'delay': delay,
                                 'coh': coh},
                                index=pos_utm.index)

        combined_data.append(pos_data)

    if mode == 'coherence' or mode == 'preavg':  # preavg is for back compat.
        # Downselect by best coherence in each axis
        coh1 = combined_data[0]['coh']
        coh1 = coh1[np.isfinite(coh1)]
        ix1 = coh1.sort_values().iloc[-ndownsel:-1].index
        coh2 = combined_data[1]['coh']
        coh2 = coh2[np.isfinite(coh2)]
        ix2 = coh2.sort_values().iloc[-ndownsel:-1].index
    elif mode == 'global_coherence':
        # Downselect to the points with the strongest overall coherence
        # An orthogonal sum of coherence for the two axes
        truecoh = combined_data[0]['coh'] ** 2 + combined_data[1]['coh'] ** 2
        truecoh = pd.DataFrame(truecoh, columns=['coh'],
                               index=combined_data[0]['coh'].index)
        ix = truecoh.sort_values('coh').iloc[-ndownsel:-1, :].index
        ix1 = ix2 = ix
    elif mode == 'distance':
        # Downselect to the closest Points
        separations = pd.DataFrame({'sep': [
            spatial.magnitude(row[1] - pos_utm.loc[ref]) for row in
            pos_utm.iterrows()]}, index=pos_utm.index)
        # Zeroth index is always "self", but needs to be included
        ix = separations.sort_values('sep').iloc[:ndownsel, :].index
        ix1 = ix2 = ix
    elif mode == 'all':
        # Just compute using all points
        ix1 = ix2 = pos_utm.index
    else:
        raise ValueError(f"Invalid mode: {mode}")

    # Average the positions in the A and B directions
    combined_data = [combined_data[0].loc[ix1], combined_data[1].loc[ix2]]
    e_a = np.mean(combined_data[0]['error_dist'])
    e_b = np.mean(combined_data[1]['error_dist'])

    # The vectors to intersect are parallel to the CMVs and have magntiude
    # equal to e_a and e_b.
    pos = spatial.compute_intersection(
        np.array([spatial.pol2rect(e_a, spatial.rect2pol(*cld_vecs[0])[1])]),
        np.array([spatial.pol2rect(e_b, spatial.rect2pol(*cld_vecs[1])[1])])
    )

    # Algorithm above assumes the origin is (0, 0). Shift by the ref pt
    pos = (pos + pos_utm.loc[ref].values).flatten()

    return pos, combined_data


def compute_delays(df, ref, navgs=5, coh_limit=0.6, freq_limit=0.02,
                   method="multi"):
    """
    Computes delay between groups of signals. Will find the delay between the
    reference and every using a transfer function between the reference and
    those possible points.

    Parameters
    ----------
    df : pandas.DataFrame
        Time series of the data. Columns are the points, rows are the time.
    ref : str
        The name of the reference point. Must be a column in df.
    navgs : int
        The number of averages to use when computing the transfer function.
        Affects both the coherence and the frequency resolution. Default is 5.
        see solarspatialtools.signalproc.averaged_tf for more information.
    coh_limit : float
        The minimum coherence required for computing the delay. Default is 0.6.
        See solarspatialtools.signalproc.tf_delay for more information.
    freq_limit : float
        The maximum frequency that will be used when computing the delay.
        Default is 0.02. See solarspatialtools.signalproc.tf_delay for more info.
    method : str
        The method for computing the delay. Options are 'fit' and 'multi'.

    Returns
    -------
    delay : np.ndarray
        The raw delay for the point pair computed from the TF phase
    coh : np.ndarray
        The average coherence for each transfer function within the window.
    """

    # Compute all TFs
    tf, tfcoh = signalproc.averaged_tf(df[ref], df, navgs=navgs, overlap=0.5,
                                       window='hamming', detrend=None)

    if method == "fit":  # A looping method
        delay = np.zeros_like(df.columns, dtype=float)
        coh = np.zeros_like(df.columns, dtype=float)
        for i, col in enumerate(tf.columns):
            tf_i = tf[col]
            coh_i = tfcoh[col]

            # Find the time delay from the TF phase
            delay[i], ix = signalproc.tf_delay(tf_i, coh_i,
                                               coh_limit=coh_limit,
                                               freq_limit=freq_limit,
                                               method='fit')

            # How good was the coherence? Average across TF
            tfsub = coh_i[tf.index < freq_limit]
            coh[i] = np.nansum(tfsub.values) / len(tfsub)

    elif method == "multi":  # The most efficient method
        delay, ix = signalproc.tf_delay(tf, tfcoh, coh_limit=coh_limit,
                                        freq_limit=freq_limit, method='multi')
        freq_ix = tf.index < freq_limit
        coh = np.nansum(tfcoh.values[freq_ix, :], axis=0) / np.nansum(freq_ix,
                                                                      axis=0)
    else:
        raise ValueError(f"Invalid method: {method}")
    return delay, coh


def remap_positions(data, remap, columns=None):
    """
    Remaps the data based on the remap indices. Each tuple is a pair of
    entity names (indices), where the first index is the original location, and
    the second index is the name of the combiner whose location is the true
    location of the entity identified in the first index.

    One common use case might be to operate on plant layout definition, where
    the index is the name of the entity and the columns are the UTM
    coordinates of the entity. This function could be used to swap entity
    positions.

    For more info on the remap input, see the `assign_positions` function.

    Parameters
    ----------
    data : pd. DataFrame
        The original data to be remapped. Its index must be the names of the
        plant entities, with the columns representing data about them. One
        common use case might be the UTM coordinates of the entities with
        columns 'E' and 'N'.
    remap : list
        The remap indices are a list of tuples, where the first element is the
        identifier of the entity, and the second element is the entity whose
        location is the true location of the first element index.
    columns : list, optional
        Specifies which columns to remap. If None, all columns are remapped.

    Returns
    -------
    data_fix : pd.DataFrame
        The remapped data frame
    """
    if columns is None:
        columns = data.columns

    # Fix the data to reflect the remapping
    data_fix = data.copy()
    for item in remap:
        for column in columns:
            data_fix.loc[item[0], column] = data[column][item[1]]
    return data_fix


def cascade_remap(remap1, remap2):
    """
    Cascades the effect of a remapping of a remapping. This function is useful
    for situations where you've remapped the data once, and then you've
    reprocessed that data under the new locations, and obtained a new
    remapping. This function will combine the two remappings into a single
    remapping that maps from the original data to the final remapped positions.

    After applying an initial remapping, the second remapping has indices that
    are now effectively scrambled by the new positions. Thus applying this in a
    sense has the effect of merging the values of the first column from the
    second remap into the second column from the first remap, and keeping the
    outer values as the result. This is a bit confusing, but the example below
    should help clarify.

    remap1 = [('A', 'B'), ('B', 'C'), ('C', 'A')]
    remap2 = [('A', 'B'), ('B', 'C'), ('C', 'A')]

    result = [('A', 'C'), ('B', 'A'), ('C', 'B')]

    Typically, values of remap inputs would be obtained from the
    `assign_positions` function.

    Parameters
    ----------
    remap1 : list of tuples
        The original remapping. Each tuple is a pair of indices, where the
        first index is the original location, and the second index is the name
        of the combiner whose location is the true location of the first index.

    remap2 : list of tuples
        The second remapping. Each tuple is a pair of indices, where the first
        index is the position within the remap1 space, and the second index is
        the new remapping also within that already remapped space.

    Returns
    -------
    out : list of tuples
        The new remapping that maps from the original locations to the new
        remapped positions after reprocessing.
    """

    # Put both into a dictionary
    A = dict(remap1)
    B = dict(remap2)

    # Create an empty dictionary to hold the outputs
    G = {}
    for key in A:
        G[key] = A[B[key]]

    out = [(k, v) for k, v in G.items()]
    out.sort(key=lambda tup: tup[0])  # sorts by 1st element in tuple
    return out


def assign_positions(original_pos, predicted_pos):
    """
    Compute the assignment solution to determine which predicted combiner
    position corresponds to which original expected combiner position. The
    assignment is done by minimizing the distance between the predicted and
    expected combiner positions.

    Understanding the outputs is a bit confusing, because we are mapping the
    expected combiner positions to the predicted combiner positions and each
    name therefore occurs twice, once in each column. So this will explain in
    excruciating detail to make sure that it's totally clear.

    The format of the the outputs is a list of tuples, where the first element
    is the name of the name of the column in the original data. The second
    element is the name of the element from the original data whose position in
    the site plan corresponds to the first element.

    For example:

        [('A', 'B'), ('B', 'C'), ('C', 'A')]

    means that the position for combiner 'B' in the original data should used
    as the true position for combiner 'A'. The position for combiner 'C' in the
    original data should be used as the true position for combiner 'B'. The
    position for combiner 'A' in the original data should be used as the true
    position for combiner 'C'.

    Stated another way, the data for 'A' indicates that those time series are
    really positioned at the ground position that was originally thought to be
    'B'.

    Parameters
    ----------
    original_pos : pd.DataFrame
        A dataframe containing the expected combiner positions. The
        dataframe should have at least two columns, the first two
        columns are assumed to be the easting and northing positions of the
        combiners in the site plan. Additional columns are ignored.

    predicted_pos : pd.DataFrame
        A dataframe containing the predicted combiner positions. The
        dataframe should have at least two columns, the first two
        columns are assumed to be the easting and northing positions of the
        combiners in the predicted site plan. Additional columns are ignored.

    Returns
    -------
    remap_indices : list of tuples
        A list of tuples, where the first element is the name of the column in
        the original data. The second element is the name of the element from
        the original data whose original position corresponds to the first
        element.

    data_out : pd.DataFrame
        A copy of the original data, but with the rows remapped to the optimal
        solution.
    """
    # If we have NaN in the data, calculations will fail. Assume that the NaN
    # values stay in place and remap the rest. This would only ever happen in
    # the predictions, because the original should come from the site plan.
    if np.any(np.isnan(predicted_pos)):
        from warnings import warn
        warn("Warning: Some NaN values in the data, analyzing only for real "
             "valued positions.")
        # Remove the nans
        sub_pred = predicted_pos.dropna()
        sub_orig = original_pos.loc[sub_pred.index]
        # Get an optimized subset without the NaNs
        subremap, _ = assign_positions(sub_orig, sub_pred)

        # Build the indices by manually adding back the NaNs.
        remap_indices = []
        for k in original_pos.index:  # Loop over all the original entries
            if np.any(np.isnan(predicted_pos.loc[k])):
                # This was an NaN case, so just remap it to itself
                remap_indices.append((k, k))
            else:
                # Find which optim. entry in subremap corresponds to this key
                for pair in subremap:
                    if pair[0] == k:
                        # Insert it
                        remap_indices.append(pair)
    else:
        # We have a good set of data, so compute the optimum

        # Construct the cost matrix, which shows the distance for each combo
        # from original to plan
        C = np.zeros((len(original_pos), len(predicted_pos)))
        E_plan = original_pos.values[:, 0]
        N_plan = original_pos.values[:, 1]
        E_delay = predicted_pos.values[:, 0]
        N_delay = predicted_pos.values[:, 1]
        for i in range(len(E_plan)):
            for j in range(len(E_plan)):
                # Error is the distances between the inferred and expected pts
                C[i, j] = np.sqrt((E_plan[i] - E_delay[j]) ** 2 +
                                  (N_plan[i] - N_delay[j]) ** 2
                                  )

        # Perform the optimization using scipy algorithm
        row_ind, col_ind = linear_sum_assignment(C)

        # Define the remapping from the original data to the new data
        remap_indices = list(zip(original_pos.index[col_ind],
                                 original_pos.index[row_ind]))
        remap_indices.sort(key=lambda tup: tup[0])

    # Create a fixed copy of the data to reflect the remapping
    data_out = original_pos.copy()
    data_out = remap_positions(data_out, remap_indices)

    return remap_indices, data_out



import matplotlib
import numpy as np


def annotate_fig(axis):
    """
    Helper to annotate a figure axis with some significant frequency timescales
    Parameters
    ----------
    axis an axis object. Get via ax = plt.gca()

    Returns
    -------
    void
    """

    annfreqs = [1 / (60*60*24), 1 / (60*60*24), np.nan,
                1 / (60 * 60 * 12), 1 / (60 * 60 * 12), np.nan,
                1 / (60 * 60), 1 / (60 * 60), np.nan,
                1 / (30 * 60), 1 / (30 * 60), np.nan,
                1 / (10 * 60), 1 / (10 * 60), np.nan,
                1 / 60, 1 / 60, np.nan,
                1 / 30, 1 / 30, np.nan,
                1 / 15, 1 / 15, np.nan,
                1 / 5, 1 / 5, np.nan]
    specfreqnames = ["1d", "12h", "1h", "30m", "10m", "1m", "30s", "15s", "5s"]
    if axis.get_xscale == matplotlib.scale.LogScale and \
            axis.get_yscale == matplotlib.scale.LinearScale:
        axis.semilogx(annfreqs,
                      np.tile([0, 0.05, 0], len(annfreqs) // 3),
                      'k-',
                      linewidth=0.5)
    elif axis.get_xscale == matplotlib.scale.LogScale and \
            axis.get_yscale == matplotlib.scale.LogScale:
        axis.loglog(annfreqs,
                    np.tile([0, 0.05, 0], len(annfreqs) // 3),
                    'k-',
                    linewidth=0.5)
    else:
        axis.plot(annfreqs,
                  np.tile([0, 0.05, 0], len(annfreqs) // 3),
                  'k-',
                  linewidth=0.5)
    for (lbl, xpos) in zip(specfreqnames, annfreqs[::3]):  # Annotate
        axis.annotate(lbl,
                      (xpos, 0.05),
                      size=8,
                      rotation='vertical',
                      ha='center',
                      va='bottom')



import numpy as np
import pandas as pd
from pyproj import Proj


def _calc_zone(lon):
    """
    Calculate the UTM zone for a given longitude, based on a 6deg window model.

    Parameters
    ----------
    lon : numeric
        The longitude in degrees

    Returns
    -------
    zone : numeric
        The UTM zone (number only, alphabetic N/S indicator excluded)
    """
    z = int((lon-(-180))/6+1)
    # minimum command accounts for +180 which otherwise produces zone 61.
    return min(z, 60)


def latlon2utm(lat, lon, zone=None):
    """
    Convert a latitude/longitude pair from WGS84 into UTM coordinates.

    Parameters
    ----------

    lat : numeric
        a value of latitude in degrees, or an indexable array of latitudes,
        or a Pandas Series (indices for lat and lon must match)

    lon : numeric
        a value of latitude in degrees, or an indexable array of latitudes,
        or a Pandas Series (indices for lat and lon must match)

    zone : numeric
        the UTM zone (number only). Calculated from longitude if None

    Returns
    -------
    (east, north) : numeric
        a tuple of the UTM (Easting, Northing) for the site. For input arrays
        or lists, the output will be a numpy array of E,N pairs. For inputs
        that are Pandas types, the output will have columns 'E' and 'N' and
        will share an index with the input.

    """

    if hasattr(lat, "__len__"):
        if not len(lat) == len(lon):
            raise ValueError("lat and lon arrays must have the same length.")
        if hasattr(lat, "index") \
                and isinstance(lat, (pd.DataFrame, pd.Series)):
            data = np.array([latlon2utm(lati, loni, zone)
                             for (lati, loni) in zip(lat, lon)])
            return pd.DataFrame(data, index=lat.index,
                                columns=["E", "N", "zone"])
        else:
            return np.array([latlon2utm(lati, loni, zone)
                             for (lati, loni) in zip(lat, lon)])
    else:
        if zone is None:
            zone = _calc_zone(lon)

        south = (lat < 0)

        p1 = Proj(proj='utm', zone=zone, ellipsis='WGS84', datum='WGS84',
                  units='m', south=south, preserve_units=True)
        (E, N) = p1(lon, lat)
        return E, N, zone


def utm2latlon(e, n, zone, south=False):
    """
    Convert from UTM coordinates to latitude/longitude (WGS84)

    Parameters
    ----------
    e : numeric
        a value of easting, or an indexable array of eastings, or a Pandas
        Series (indices for e and n must match)

    n : numeric
        a value of northing, or an indexable array of northings, or a Pandas
        Series (indices for e and n must match)

    zone : numeric
        the UTM zone, which is required for this conversion. Zone should
        include only the number (i.e. 32U should be listed as 32).

    south : bool
        False for northern hemisphere, True for southern hemisphere

    Returns
    -------
    (lat, lon) : numeric
        tuple containing latitude, longitude in degrees. For input arrays
        or lists, the output will be a numpy array of lat,lon pairs. For inputs
        that are Pandas types, the output will have columns 'lat' and 'lon' and
        will share an index with the input.
    """

    if hasattr(e, "__len__"):
        if not len(e) == len(n):
            raise ValueError("e and n arrays must have the same length.")
        if hasattr(e, "index") \
                and isinstance(e, (pd.DataFrame, pd.Series)):
            data = [utm2latlon(ei, ni, zone, south) for (ei, ni) in zip(e, n)]
            return pd.DataFrame(data, index=e.index, columns=['lat', 'lon'])
        else:
            return [utm2latlon(ei, ni, zone, south) for (ei, ni) in zip(e, n)]
    else:
        p1 = Proj(proj='utm', zone=zone, ellipsis='WGS84', datum='WGS84',
                  units='m', south=south)
        (lon, lat) = p1(e, n, inverse=True)
        return lat, lon


def latlon2lcs(lat, lon, origin_lat, origin_lon, zone=None):
    """
    Convert from lat/lon to a referenced local coordinate system (LCS) for
    wobas. LCS uses a UTM projection but offsets the values to a given origin.

    Parameters
    ----------
    lat : numeric
        target point latitude in deg

    lon : numeric
        target point longitude in deg

    origin_lat : numeric
        The latitude of the LCS origin in degrees

    origin_lon : numeric
        The longitude of the LCS origin in degrees

    zone : numeric
        The UTM zone (obtained from origin longitude if not specified)

    Returns
    -------
    (east, north) : numeric
        The Easting and Northing of the target point in the LCS coordinates
    """
    # Convert both coordinates to UTM
    e_i, n_o, _ = latlon2utm(origin_lat, origin_lon, zone=zone)
    e_t, n_t, _ = latlon2utm(lat, lon, zone=zone)

    # Shift by origin to convert to LCS
    # (LCS is UTM, re-zeroed to the origin's coordinates)
    e_t -= e_i
    n_t -= n_o

    return e_t, n_t


def lcs2latlon(east, north, origin_lat, origin_lon, zone=None):
    """
    Convert from the origin referenced local coordinate system (LCS) for wobas
    back to a lat/lon coordinate system. LCS uses a UTM projection but offsets
    the values to a given origin.
    Parameters
    ----------
    east : numeric
        The easting of the point in the LCS

    north : numeric
        The northing of the point in the LCS

    origin_lat : numeric
        The latitude of the LCS origin in degrees

    origin_lon : numeric
        The longitude of the LCS origin in degrees

    zone : numeric
        The UTM zone (obtained from origin longitude if not specified)

    Returns
    -------
    (lat, lon) : numeric
        Tuple containing the latitude and longitude of the point in degrees
    """
    # Calculate the zone and south flag
    if zone is None:
        zone = _calc_zone(origin_lon)
    south = origin_lat < 0

    # Get the origin in UTM
    e_o, n_o, _ = latlon2utm(origin_lat, origin_lon)

    # Shift the target
    e_t = e_o + east
    n_t = n_o + north

    # Calculate
    lat, lon = utm2latlon(e_t, n_t, zone, south)
    return lat, lon


def dot(vec_a, vec_b):
    """
    Vector dot product for 2-D cartesian

    Parameters
    ----------
    vec_a : numeric
        the 2-D cartesian vector as an (x, y) indexable

    vec_b : numeric
        the 2-D cartesian vector as an (x, y) indexable

    Returns
    -------
    A dot B : numeric
        the dot product of the two vectors
    """
    return vec_a[0] * vec_b[0] + vec_a[1] * vec_b[1]


def unit(vec):
    """
    Get the unit vector matching a vector's direction

    Parameters
    ----------
    vec : numeric
        the 2-D cartesian vector as an (x, y) indexable

    Returns
    -------
    (xi, yi) : numeric
        the 2-D cartesian vector scaled to a magnitude of 1
    """
    return vec / magnitude(vec)


def magnitude(vec):
    """
    Get the magnitude of a vector

    Parameters
    ----------
    vec : numeric
        the 2-D cartesian vector as an (x, y) indexable

    Returns
    -------
    magnitude : numeric
        the magnitude of the vector
    """
    # respond differently for Pandas Types
    if isinstance(vec, pd.Series):
        return np.sqrt(vec.iloc[0]**2 + vec.iloc[1]**2)
    return np.sqrt(vec[0]**2+vec[1]**2)


def rect2pol(x, y):
    """
    2D cartesian vector to polar form

    Parameters
    ----------
    x : numeric
        the x position

    y : numeric
        the y position

    Returns
    -------
    (r, theta) : numeric
        the polar form (r, theta) with theta in +/- pi radians
    """
    r = magnitude([x, y])
    theta = np.arctan2(y, x)
    return r, theta


def pol2rect(r, theta):
    """
    2D polar vector to cartesian form

    Parameters
    ----------
    r : numeric
        the radius coordinate

    theta : numeric
        the angle coordinate in radians

    Returns
    -------
    (x, y) : numeric
        a tuple of the vector
    """

    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y


def rotate_vector(vector, theta):
    """
    2D rotation of a vector in cartesian form

    Parameters
    ----------
    vector : (x, y) numeric
        A tuple (or numpy array) containing the input vector.
        To operate on multiple points, vector should be of the form:
        ((x1, x2, x3, x4), (y1, y2, y3, y4))

    theta : numeric
        Angle of rotation in radians

    Returns
    -------
    (x, y) : numeric
        A tuple (or numpy array) containing the rotated vector. Matches data
        type of input. If input an array of x's and y's, output will be
        oriented ((x1', x2', x3', x4'), (y1', y2', y3', y4'))
    """

    rot_matrix = np.array([[np.cos(theta), -np.sin(theta)],
                           [np.sin(theta), np.cos(theta)]])
    newvec = np.matmul(rot_matrix, vector)

    if type(vector) in [type(()), type([])]:
        return tuple(newvec)
    else:
        return newvec


def project_vectors(vectors, dir_vector):
    """
    Convert a set of vectors into a list of distances in a given direction.
    See also: compute_vectors for computing the vector array. Units will match
    those provided for the input vectors. Should probably be done in a
    rectilinear system like UTM.

    Parameters
    ----------
    vectors : numeric
        iterable set of vectors pointing from a reference to sites of interest.

    dir_vector : numeric
        tuple of (x, y) showing the cartesian direction that the vectors should
        be projected onto. Length of vector greater than 1 will be ignored
        (i.e. will be converted to unit vector as the projection is done).

    Returns
    -------
    dot_positions : numeric
        The dot product of each vector with the projection vector. Essentially
        represents the length of each vector in the given directions, with
        negative numbers implying opposite pointing. If the input is a Pandas
        type, the output will be in a DataFrame with an index that matches the
        input and has columns 'dist'.
    """

    if hasattr(vectors, "index") \
            and isinstance(vectors, (pd.DataFrame, pd.Series)):
        pandas_mode = True
        vec_array = np.array(vectors)
    else:
        pandas_mode = False
        vec_array = vectors

    unit_dir = unit(dir_vector)
    v = np.array(vec_array)
    dot_positions = v[:, 0] * unit_dir[0] + v[:, 1] * unit_dir[1]

    if pandas_mode:
        return pd.DataFrame(dot_positions,
                            index=vectors.index, columns=['dist'])
    else:
        return dot_positions


def compute_vectors(x, y, refpt):
    """
    Compute a list of vectors pointing from a reference origin to each site.
    Units will match those given by x, y, which means that ideally this
    function would be called with inputs in a rectilinear system such as UTM.

    Parameters
    ----------
    x : numeric
        An indexable array of x coordinates

    y : numeric
        An indexable array of y coordinates

    refpt : numeric
        The reference position as a tuple (x, y) or other indexable pair

    Returns
    -------
    vectors : numeric
        An ordered list of vectors (x,y) from ref to each point in meters. If
        the input is a Pandas type, the output will be in a DataFrame with an
        index that matches the input and has columns 'dx' and 'dy'.

    """

    # Find ref
    pt0 = np.array(refpt)[0:2]  # Make sure it doesn't contain extra columns

    # Calculate each point's vector
    vectors = []
    for x_i, y_i in zip(x, y):
        pti = np.array((x_i, y_i))

        # Compute the vector
        ptdiff = pti - pt0
        vectors.append(ptdiff)

    if hasattr(x, "index") and isinstance(x, (pd.DataFrame, pd.Series)):
        return pd.DataFrame(np.array(vectors),
                            index=x.index, columns=['dx', 'dy'])
    else:
        return np.array(vectors)


def compute_intersection(A, B):
    """
    Computes intersection, C, of the lines perpendicular to the tips of two
    vectors (A and B) who share a starting point at the origin.

    Vectors are specified in nx2 arrays and computations of Ci are made for
    each vector pair Ai, Bi.

    Parameters
    ----------
    A : np.array
        nx2 array of the vectors. Outer index is the vector instance, inner
        index is the coordinate x,y:
        [[ax1,ay1], [ax2,ay2], ..., [axn,ayn]]
    B : np.array
        nx2 array of the vectors. Outer index is the vector instance, inner
        index is the coordinate x,y:
        [[bx1,by1], [bx2,by2], ..., [bxn,byn]]

    Returns
    -------
    C : np.array
        nx2 array of computed positions of the intersection.
        [[cx1,cy1], [cx2,cy2], ..., [cxn,cyn]]
    """

    # Alternate method available via line intersections at:
    # https://stackoverflow.com/questions/3252194/numpy-and-line-intersections

    # Derive by computing unit vectors a & b in dir of A&B
    # Then compute a dot C and b dot C. These should be equal to the
    # magnitudes of A and B. You now have an equation for cx, cy in terms
    # of the known a, b, A, B. Thanks @ Nate Williams and David Starling.

    Ax = A[:, 0]
    Ay = A[:, 1]
    Bx = B[:, 0]
    By = B[:, 1]
    Amag2 = Ax**2 + Ay**2  # magnitude squared
    Bmag2 = Bx**2 + By**2
    cx = (Bmag2 * Ay - By * Amag2) / (Bx*Ay-Ax*By)
    cy = (Amag2 * Bx - Ax * Bmag2) / (Bx*Ay-Ax*By)

    C = np.array([cx, cy]).T

    return C



import pytest
from pytest import approx
import numpy as np
from numpy.testing import assert_allclose
import pandas as pd

from solarspatialtools import clearsky_index


def test_clearsky_index():
    ghi = np.array([-1., 0., 1., 500., 1000., np.nan])
    ghi_measured, ghi_modeled = np.meshgrid(ghi, ghi)
    # default max_clearsky_index
    with np.errstate(invalid='ignore', divide='ignore'):
        out = clearsky_index(ghi_measured, ghi_modeled)
    expected = np.array(
        [[1.    , 0.    , 0.    , 0.    , 0.    , np.nan],
         [0.    , 0.    , 0.    , 0.    , 0.    , np.nan],
         [0.    , 0.    , 1.    , 2.    , 2.    , np.nan],
         [0.    , 0.    , 0.002 , 1.    , 2.    , np.nan],
         [0.    , 0.    , 0.001 , 0.5   , 1.    , np.nan],
         [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]])
    assert_allclose(out, expected, atol=0.001)
    # specify max_clearsky_index
    with np.errstate(invalid='ignore', divide='ignore'):
        out = clearsky_index(ghi_measured, ghi_modeled, max_clearsky_index=1.5)
    expected = np.array(
        [[1.    , 0.    , 0.    , 0.    , 0.    , np.nan],
         [0.    , 0.    , 0.    , 0.    , 0.    , np.nan],
         [0.    , 0.    , 1.    , 1.5   , 1.5   , np.nan],
         [0.    , 0.    , 0.002 , 1.    , 1.5   , np.nan],
         [0.    , 0.    , 0.001 , 0.5   , 1.    , np.nan],
         [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]])
    assert_allclose(out, expected, atol=0.001)

    # GHI series & CS series
    times = pd.date_range(start='20180601', periods=2, freq='12h')
    ghi_measured = pd.Series([100,  500], index=times)
    ghi_modeled = pd.Series([500, 1000], index=times)
    out = clearsky_index(ghi_measured, ghi_modeled)
    expected = pd.Series([0.2, 0.5], index=times)
    assert isinstance(out, pd.Series)
    assert (out.index == times).all()
    assert out.values == approx(expected.values)

    # GHI 2D frame & CS 1D series
    times = pd.date_range(start='20180601', periods=3, freq='12h')
    ghi_measured = pd.DataFrame([[100,  100], [200, 200], [500, 500]], index=times)
    ghi_modeled = pd.Series([500, 800, 1000], index=times)
    out = clearsky_index(ghi_measured, ghi_modeled)
    expected = pd.DataFrame([[0.2, 0.2], [0.25, 0.25], [0.5, 0.5]], index=times)
    assert isinstance(out, pd.DataFrame)
    assert (out.index == times).all()
    assert out.values == approx(expected.values)



import pytest
from pytest import approx, raises
import pandas as pd
import numpy as np
from scipy import signal
from scipy.interpolate import interp1d
from solarspatialtools.signalproc import averaged_psd, averaged_tf, interp_tf, tf_delay, xcorr_delay, apply_delay, correlation, compute_delays, _fftcorrelate


@pytest.fixture(params=[0, 0.2, -0.2, 0.4, -0.4, 1, -1])
def corr_data(request):
    dt = 0.01
    dly = request.param  # variable delay
    t = np.arange(0, 10, 0.01)
    f = 1

    # Single half-period pulse of sine wave
    x1 = np.sin(2 * np.pi * f * t)
    x1[t < 4*(1/f)] = 0
    x1[t > 4.5*(1/f)] = 0

    # Delayed version
    x2 = np.sin(2 * np.pi * f * (t - dly))  # delay by 0.2 seconds
    x2[t < 4*(1/f) + dly] = 0
    x2[t > 4.5*(1/f) + dly] = 0
    return dt, t, x1, x2, dly


@pytest.fixture(params=['energy', 'coeff', 'unbiased_energy', 'unbiased_coeff'])
def scaling(request):
    return request.param


def test_correlation_identity(corr_data, scaling):
    dt, t, x1, x2, dly = corr_data
    c, lag = correlation(x1, x1, scaling=scaling)
    imax = np.argmax(c)

    assert c[imax] == approx(1)
    assert -lag[imax]*(t[1]-t[0]) == approx(0)


def test_correlation_shift(corr_data, scaling):
    d, t, x1, x2, dly = corr_data
    c, lag = correlation(x1, x2, scaling=scaling)
    imax = np.argmax(c)
    assert -lag[imax]*(t[1]-t[0]) == approx(dly)  # -lag * dt == t_shift


def test_correlation_illegal(corr_data):
    d, t, x1, x2, dly = corr_data
    with raises(ValueError):
        c, lag = correlation(x1, x2, scaling="illegal")


@pytest.mark.parametrize("scaling", ['coeff', 'none'])
def test_fftcorrelate_identity(corr_data, scaling):
    dt, t, x1, x2, dly = corr_data
    c, l = _fftcorrelate(x1, x1, scaling)
    cr, lag = correlation(x1, x1, scaling=scaling)
    assert np.allclose(c, cr)
    assert np.allclose(l, lag)


@pytest.mark.parametrize("scaling", ['coeff', 'none'])
def test_fftcorrelate_shift(corr_data, scaling):
    d, t, x1, x2, dly = corr_data
    c, l = _fftcorrelate(x1, x2, scaling)
    cr, lag = correlation(x1, x2, scaling=scaling)
    assert np.allclose(c, cr)
    assert np.allclose(l, lag)


def test_averaged_psd():
    # Create a simple sinusoidal signal
    fs = 10  # sample rate
    T = 5.0    # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    x = 0.5*np.sin(2*np.pi*2*t)  # +0.1*np.sin(2*np.pi*10*t)
    input_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))

    # Calculate the averaged PSD using the function
    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = 'linear'
    scaling = 'density'
    psd = averaged_psd(input_tsig, navgs, overlap, window, detrend, scaling)

    # Calculate the PSD directly using scipy.signal.welch
    freqs, psd_direct = signal.welch(x, fs, window, nperseg=len(x)//navgs,
                                     noverlap=int(overlap*len(x)//navgs),
                                     detrend=detrend, scaling=scaling)

    # Check that the PSDs match
    assert np.allclose(psd, pd.DataFrame(psd_direct), atol=1e-5)


def test_averaged_tf():
    # Create a simple sinusoidal signal
    fs = 10  # sample rate
    T = 5.0 # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    x = 0.5*np.sin(2*np.pi*2*t)  # input signal
    y = 0.5*np.sin(2*np.pi*2*t + np.pi/4)  # output signal with phase shift
    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    y_tsig = pd.Series(y, index=pd.to_timedelta(t, 's'))

    # Calculate the averaged transfer function using the function
    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = 'linear'
    tf, coh = averaged_tf(x_tsig, y_tsig, navgs, overlap, window, detrend)

    # Calculate the transfer function directly using scipy csd and welch
    freqs, Pxy = signal.csd(x, y, fs, window, nperseg=len(x)//navgs,
                            noverlap=int(overlap*len(x)//navgs),
                            detrend=detrend)
    freqs, Pxx = signal.welch(x, fs, window, nperseg=len(x)//navgs,
                              noverlap=int(overlap*len(x)//navgs),
                              detrend=detrend)
    freqs, Pyy = signal.welch(y, fs, window, nperseg=len(x) // navgs,
                              noverlap=int(overlap * len(x) // navgs),
                              detrend=detrend)

    tf_direct = Pxy / Pxx
    coh_direct = np.abs(Pxy)**2 / (Pxx * Pyy)
    # freqs, coh_direct = signal.coherence(x, y, fs, window, nperseg=len(x)//navgs,
    #                               noverlap=int(overlap*len(x)//navgs),
    #                               detrend=detrend)

    # Check that the transfer functions match
    assert np.allclose(tf, pd.DataFrame(tf_direct), atol=1e-5)
    assert np.allclose(coh, pd.DataFrame(coh_direct), atol=1e-5)
    assert np.allclose(tf.index, freqs, atol=1e-5)


def test_interp_tf():
    # Create a simple transfer function
    freqs = np.linspace(0, 1, 100)
    mag = np.abs(np.sin(2*np.pi*freqs))
    phase = np.unwrap(np.sin(2*np.pi*freqs)+freqs*6)
    tf = pd.Series(mag * np.exp(1j * phase), index=freqs)

    # Calculate the interpolated transfer function using the function
    new_freq = np.linspace(0, 1, 200)
    interp_tf_result = interp_tf(new_freq, tf)

    # Calculate the interpolated transfer function directly
    interp_mag_func = interp1d(freqs, mag, axis=0)
    interp_phase_func = interp1d(freqs, phase, axis=0)
    interp_mag_direct = interp_mag_func(new_freq)
    interp_phase_direct = interp_phase_func(new_freq)
    interp_tf_direct = interp_mag_direct * np.exp(1j * interp_phase_direct)

    # Check that the interpolated transfer functions match
    assert np.allclose(interp_tf_result, interp_tf_direct, atol=1e-5)


@pytest.mark.parametrize("delay", [-5.0, -2.5, -0.5, 0.0, 0.5, 2.5, 5.0])
def test_tf_delay(delay):
    # delay = 0.05  # delay in seconds
    fs = 100  # sample rate
    T = 500.0  # total seconds
    np.random.seed(2023)
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    noise = np.random.random(len(t))/5  # add broadband noise
    x = 0.5*np.sin(2*np.pi*2*t) + noise  # noisy signal
    y = np.roll(x, int(delay*fs))  # Explicitly delay the signal
    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    y_tsig = pd.Series(y, index=pd.to_timedelta(t, 's'))

    # Calculate the averaged transfer function using the function
    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = 'linear'
    tf, coh = averaged_tf(x_tsig, y_tsig, navgs, overlap, window, detrend)

    # Calculate the delay using the function
    delay_result, _ = tf_delay(tf, coh, coh_limit=0.6, freq_limit=1, method='multi')

    # Check that the delay matches the expected delay
    assert np.isclose(delay_result, delay, atol=1e-2)


@pytest.mark.parametrize("delay", [-5.0, -2.5, -0.5, 0.0, 0.5, 2.5, 5.0])
def test_tf_delay_multi(delay):
    np.random.seed(2023)
    # Create a simple sinusoidal signal
    fs = 500  # sample rate
    T = 1000.0  # seconds
    t = np.linspace(0, T, int(T * fs), endpoint=False)  # time variable

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    y1 = np.roll(x, int(delay * fs))
    y2 = np.roll(x, int(2 * delay * fs))
    y3 = np.roll(x, int(3 * delay * fs))
    y4 = np.roll(x, int(4 * delay * fs))
    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    ysigs = [pd.Series(y, index=pd.to_timedelta(t, 's')) for y in
             [y1, y2, y3, y4]]
    ysigs_df = pd.DataFrame(np.array(ysigs).T, columns=[0, 1, 2, 3],
                            index=ysigs[0].index)

    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = None

    tf, coh = averaged_tf(x_tsig, ysigs_df, navgs, overlap, window, detrend)

    # Calculate the delay using the function
    delay_result, _ = tf_delay(tf, coh, coh_limit=0.6, freq_limit=1, method='multi')

    # Check that the delay matches the expected delay
    delay_expect = np.array([delay, 2*delay, 3*delay, 4*delay])
    assert np.allclose(delay_result, delay_expect, atol=1e-2)


def test_tf_delay_nan():
    delay = 5
    np.random.seed(2023)
    # Create a simple sinusoidal signal
    fs = 500  # sample rate
    T = 1000.0  # seconds
    t = np.linspace(0, T, int(T * fs), endpoint=False)  # time variable

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    y1 = np.roll(x, int(delay * fs))
    y2 = np.roll(x, int(2 * delay * fs))
    y3 = np.roll(x, int(3 * delay * fs))
    y4 = np.nan * np.zeros_like(y3)

    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    ysigs = [pd.Series(y, index=pd.to_timedelta(t, 's')) for y in
             [y1, y2, y3, y4]]
    ysigs_df = pd.DataFrame(np.array(ysigs).T, columns=[0, 1, 2, 3],
                            index=ysigs[0].index)

    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = None

    tf, coh = averaged_tf(x_tsig, ysigs_df, navgs, overlap, window, detrend)

    # Calculate the delay using the function
    delay_result, _ = tf_delay(tf, coh, coh_limit=0.6, freq_limit=1, method='multi')

    # Check that the delay matches the expected delay
    delay_expect = np.array([delay, 2*delay, 3*delay, np.nan])
    assert np.allclose(delay_result[0:-1], delay_expect[0:-1], atol=1e-2)
    assert np.isnan(delay_result[-1])


@pytest.mark.parametrize("delay", [-5.0, -2.5, -0.5, 0.0, 0.5, 2.5, 5.0])
def test_xcorr_delay(delay):
    # delay = -0.05  # delay in seconds
    fs = 100  # sample rate
    T = 500.0  # total seconds
    np.random.seed(2023)
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    noise = np.random.random(len(t))/5  # add broadband noise
    x = 0.5*np.sin(2*np.pi*0.01*t) + noise  # noisy signal
    y = np.roll(x, int(delay*fs))  # Explicitly delay the signal
    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    y_tsig = pd.Series(y, index=pd.to_timedelta(t, 's'))

    # Calculate the averaged transfer function using the function
    delay_result, _ = xcorr_delay(x_tsig, y_tsig, "coeff")

    # Check that the delay matches the expected delay
    assert np.isclose(delay_result, delay, atol=1e-2)


@pytest.mark.parametrize("delay", [-5.0, -2.5, -0.5, 0.0, 0.5, 2.5, 5.0])
def test_apply_delay(delay):
    # Create a simple transfer function
    freqs = np.linspace(0.01, 1, 100)  # frequencies
    phase = -2 * np.pi * freqs * delay  # phase delay
    mag = np.ones_like(freqs)  # magnitude
    coh = np.ones_like(freqs)  # coherence
    tf = pd.DataFrame({'tf': mag * np.exp(1j * phase)}, index=freqs)

    # Apply the delay using the function
    tf_delayed = apply_delay(tf, delay)

    # Calculate the expected delayed transfer function
    tf_expected = tf.copy().values.flatten()
    tf_expected = tf_expected * np.exp(2 * np.pi * 1j * tf.index * -delay)

    # Check that the delayed transfer function matches the expected one
    assert np.allclose(tf_delayed, tf_expected, atol=1e-5)


def test_averaged_psd_multi():
    np.random.seed(2023)
    # Create a simple sinusoidal signal
    fs = 10  # sample rate
    T = 5.0    # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable

    x1 = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))
    x2 = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))
    x3 = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))
    x4 = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    xs = [x1, x2, x3, x4]
    tsigs = [pd.Series(x, index=pd.to_timedelta(t, 's')) for x in xs]
    tsigs_df = pd.DataFrame(np.array(tsigs).T, columns=[0, 1, 2, 3], index=tsigs[0].index)
    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = 'linear'
    scaling = 'density'
    psds = averaged_psd(tsigs_df, navgs, overlap, window, detrend, scaling)

    # Calculate the PSD directly using scipy.signal.welch
    freqs, psd_direct = signal.welch(xs, fs, window, nperseg=len(x1)//navgs,
                                     noverlap=int(overlap*len(x1)//navgs),
                                     detrend=detrend, scaling=scaling)

    # Check that the PSDs match
    assert np.allclose(psds, pd.DataFrame(psd_direct.T), atol=1e-5)


def test_averaged_tf_multiout():
    np.random.seed(2023)
    # Create a simple sinusoidal signal
    fs = 500  # sample rate
    T = 1000.0    # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    delay = 3

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    y1 = np.roll(x, int(delay*fs))
    y2 = np.roll(x, int(2*delay * fs))
    y3 = np.roll(x, int(3*delay * fs))
    y4 = np.roll(x, int(4*delay * fs))
    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    ysigs = [pd.Series(y, index=pd.to_timedelta(t, 's')) for y in [y1, y2, y3, y4]]
    ysigs_df = pd.DataFrame(np.array(ysigs).T, columns=[0, 1, 2, 3], index=ysigs[0].index)

    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = None

    tf, coh = averaged_tf(x_tsig, ysigs_df, navgs, overlap, window, detrend)

    loop_dat = [averaged_tf(x_tsig, ysig, navgs, overlap, window, detrend) for ysig in ysigs]
    tf_loop = [dat[0] for dat in loop_dat]
    coh_loop = [dat[1] for dat in loop_dat]
    tf_loop = np.array(tf_loop)[:, :, 0].T
    coh_loop = np.array(coh_loop)[:, :, 0].T

    # Check that the PSDs match
    assert np.allclose(tf, tf_loop, atol=1e-5)
    assert np.allclose(coh, coh_loop, atol=1e-5)


def test_averaged_tf_multiin():
    np.random.seed(2023)
    # Create a simple sinusoidal signal
    fs = 500  # sample rate
    T = 1000.0    # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    delay = 3

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    y1 = np.roll(x, int(delay*fs))
    y2 = np.roll(x, int(2*delay * fs))
    y3 = np.roll(x, int(3*delay * fs))
    y4 = np.roll(x, int(4*delay * fs))
    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    ysigs = [pd.Series(y, index=pd.to_timedelta(t, 's')) for y in [y1, y2, y3, y4]]
    ysigs_df = pd.DataFrame(np.array(ysigs[0]).T, columns=[0], index=ysigs[0].index)

    xsigs = [x_tsig, ysigs[0], ysigs[1], ysigs[2]]
    xsigs_df = pd.DataFrame(np.array(xsigs).T, columns=[0, 1, 2, 3], index=ysigs[0].index)
    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = None

    tf, coh = averaged_tf(xsigs_df, ysigs_df, navgs, overlap, window, detrend)

    loop_dat = [averaged_tf(xsig, ysigs[0], navgs, overlap, window, detrend) for xsig in xsigs]
    tf_loop = [dat[0] for dat in loop_dat]
    coh_loop = [dat[1] for dat in loop_dat]
    tf_loop = np.array(tf_loop)[:, :, 0].T
    coh_loop = np.array(coh_loop)[:, :, 0].T

    # Check that the PSDs match
    assert np.allclose(tf, tf_loop, atol=1e-5)
    assert np.allclose(coh, coh_loop, atol=1e-5)


def test_averaged_tf_multiboth():
    np.random.seed(2023)
    # Create a simple sinusoidal signal
    fs = 500  # sample rate
    T = 1000.0    # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    delay = 3

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    y1 = np.roll(x, int(delay*fs))
    y2 = np.roll(x, int(2*delay * fs))
    y3 = np.roll(x, int(3*delay * fs))
    y4 = np.roll(x, int(4*delay * fs))
    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    ysigs = [pd.Series(y, index=pd.to_timedelta(t, 's')) for y in [y1, y2, y3, y4]]
    ysigs_df = pd.DataFrame(np.array(ysigs).T, columns=[0, 1, 2, 3], index=ysigs[0].index)

    xsigs = [x_tsig, ysigs[0], ysigs[1], ysigs[2]]
    xsigs_df = pd.DataFrame(np.array(xsigs).T, columns=[0, 1, 2, 3], index=ysigs[0].index)
    navgs = 5
    overlap = 0.5
    window = 'hamming'
    detrend = None

    tf, coh = averaged_tf(xsigs_df, ysigs_df, navgs, overlap, window, detrend)

    loop_dat = [averaged_tf(xsig, ysig, navgs, overlap, window, detrend) for xsig, ysig in zip(xsigs, ysigs)]
    tf_loop = [dat[0] for dat in loop_dat]
    coh_loop = [dat[1] for dat in loop_dat]
    tf_loop = np.array(tf_loop)[:, :, 0].T
    coh_loop = np.array(coh_loop)[:, :, 0].T

    # Check that the PSDs match
    assert np.allclose(tf, tf_loop, atol=1e-5)
    assert np.allclose(coh, coh_loop, atol=1e-5)

@pytest.fixture(params=['loop', 'fft'])
def compute_delays_modes(request):
    return request.param


@pytest.mark.parametrize("delay", [-200, -50.0, -25, -5, 0.0, 5, 25, 50, 200])
def test_compute_delays(delay, compute_delays_modes):
    np.random.seed(2023)
    # Create a simple sinusoidal signal
    fs = 100  # sample rate
    T = 500.0  # seconds
    t = np.linspace(0, T, int(T * fs), endpoint=False)  # time variable
    # delay = 0.25

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + 5*np.random.random(len(t))

    ys = [x]
    delay_ins = [0]
    n = 50
    for i in range(n):
        deli = delay / n * i
        yi = np.roll(x, int(deli * fs))
        ys.append(yi)
        delay_ins.append(deli)

    x_tsig = pd.Series(x, index=pd.to_timedelta(t, 's'))
    ysigs = [pd.Series(y, index=pd.to_timedelta(t, 's')) for y in ys]
    xsigs = [x_tsig for y in ys]
    xsigs = pd.DataFrame(np.array(xsigs).T, index=x_tsig.index)

    delays, _ = compute_delays(xsigs, ysigs, compute_delays_modes)

    assert np.allclose(delays, delay_ins)
